{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  :: QuickReport 3.0 for Delphi and C++Builder               ::
  ::                                                         ::
  :: convmain.PAS - MAIN UNIT for Qr1 to QR3 converter tool  ::
  ::                                                         ::
  :: Copyright (c) 1999 QuSoft AS                            ::
  :: All Rights Reserved                                     ::
  ::                                                         ::
  :: web: http://www.qusoft.com                              ::
  ::                                                         ::
  ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

unit convmain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Mask, ExtCtrls, ComCtrls, Buttons;

resourcestring
  STR_APP_KEY       = '\Software\QuSoft\Convert';
  STR_SRC_KEY       = 'SourceFile';
  STR_DEST_KEY      = 'DestinationFile';
  STR_DFM_KEY       = 'DfmAsText';
  STR_OBJECT        = 'OBJECT ';
  STR_DELETE_MARKER = '//DeLeTeQR ';
  STR_FILE_EXISTS   = 'The file %s already exists.  Do you want to overwrite it?';

const
  QR_Band       = 1;
  QR_Printable  = 2;
  QR_Datasource = 3;
  QR_dataset    = 4;
  QR_Report     = 5;
  QR_DBCALC     = 6;
  QR_DetailLink = 7;
  QR_Image      = 8;
  QR_group      = 9;

type
  TEventRec = record
    Event : string;
    EventName : string;
    Before,
    After: string;
  end;

const
  // This array holds the report events that need to have the parameter list updated
  QR_Events: array[0..3] of TEventRec = (
    (Event: 'BeforePrint';
      Before : '(var PrintReport: Boolean)'; After: '(Sender: TCustomQuickRep; var PrintReport: Boolean)';),
    (Event: 'OnEndPage';
      Before : '(Sender: TObject)'; After: '(Sender: TCustomQuickRep)';),
    (Event: 'OnNeedData';
      Before : '(var MoreData: Boolean)'; After: '(Sender: TObject; var MoreData: Boolean)';),
    (Event: 'OnStartPage';
      Before : '(Sender: TObject)'; After: '(Sender: TCustomQuickRep)';)
  );

  // This array contains the list of report properties that will
  // transfer over to QR3.
  // The margins and paper dimensions are not touched.  You will
  // have to set these properties by hand.
  qr_Properties: array[0..17] of TEventRec = (
    (Before: 'OnNeedData';              After: 'OnNeedData';),
    (Before: 'AfterPrint';              After: 'AfterPrint';),
    (Before: 'BeforePrint';             After: 'BeforePrint';),
    (Before: 'OnEndPage';               After: 'OnEndPage';),
    (Before: 'OnStartPage';             After: 'OnStartPage';),
    (Before: 'DataSource';              After: 'Dataset';),
    (Before: 'ShowProgress';            After: 'ShowProgress';),
    (Before: 'ReportTitle';             After: 'ReportTitle';),
    (Before: 'Columns';                 After: 'Page.Columns';),
    (Before: 'Orientation';             After: 'Page.Orientation';),
    (Before: 'PaperSize';               After: 'Page.PaperSize';),
    (Before: 'PageFrame.Color';         After: 'Frame.Color';),
    (Before: 'PageFrame.DrawBottom';    After: 'Frame.DrawBottom';),
    (Before: 'PageFrame.DrawLeft';      After: 'Frame.DrawLeft';),
    (Before: 'PageFrame.DrawRight';     After: 'Frame.DrawRight';),
    (Before: 'PageFrame.DrawTop';       After: 'Frame.DrawTop';),
    (Before: 'PageFrame.Style';         After: 'Frame.Style';),
    (Before: 'PageFrame.Width';         After: 'Frame.Width';)
  );

  // This array contains the list of band properties that will
  // transfer over to QR3.  Propeties not on this list will not
  // be transferred
  Band_Properties: array[0..9] of TEventRec = (
    (Before: 'Height';                  After: 'Height';),
    (Before: 'BandType';                After: 'BandType';),
    (Before: 'Color';                   After: 'Color';),
    (Before: 'ForceNewPage';            After: 'ForceNewPage';),
    (Before: 'Frame.DrawTop';           After: 'Frame.DrawTop';),
    (Before: 'Frame.DrawBottom';        After: 'Frame.DrawBottom';),
    (Before: 'Frame.DrawLeft';          After: 'Frame.DrawLeft';),
    (Before: 'Frame.DrawRight';         After: 'Frame.DrawRight';),
    (Before: 'PageFrame.Style';         After: 'Frame.Style';),
    (Before: 'PageFrame.Width';         After: 'Frame.Width';)
  );

type
  TQRConvert = class
    srcDFM,
    srcPAS,
    DestDFM,
    DestPAS: tstringlist;
    constructor Create;
    destructor Destroy; override;
  end;

  TfrmConvert = class(TForm)
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    TabSheet3: TTabSheet;
    Panel4: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    btnConvert: TButton;
    Panel5: TPanel;
    Panel1: TPanel;
    memoDestDfm: TMemo;
    Panel6: TPanel;
    Panel7: TPanel;
    memoDestPas: TMemo;
    Splitter2: TSplitter;
    Splitter3: TSplitter;
    Panel8: TPanel;
    Panel3: TPanel;
    memoSrcDFM: TMemo;
    Panel9: TPanel;
    Panel10: TPanel;
    memoSrcPas: TMemo;
    Splitter1: TSplitter;
    Panel11: TPanel;
    Panel12: TPanel;
    Panel2: TPanel;
    tvQR: TTreeView;
    Splitter4: TSplitter;
    Panel13: TPanel;
    Memo2: TMemo;
    chkText: TCheckBox;
    SrcEdit: TEdit;
    btnSrc: TBitBtn;
    odSrc: TOpenDialog;
    DestEdit: TEdit;
    btnDest: TBitBtn;
    sdDest: TSaveDialog;
    ProgressBar: TProgressBar;
    StatusBar: TStatusBar;
    procedure btnConvertClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure tvQRChange(Sender: TObject; Node: TTreeNode);
    procedure btnSrcClick(Sender: TObject);
    procedure btnDestClick(Sender: TObject);
    procedure StatusBarDrawPanel(StatusBar: TStatusBar;
      Panel: TStatusPanel; const Rect: TRect);
  private
    { Private declarations }
    QRConvert: TQRConvert;
    SrcDFM,
    SrcPAS,
    DestDFM,
    DestPas: string;
    QRHeaderBands,
    QRImages,
    QRSubDetails,
    QRDataSources: TStringList;
    ReportNode: TTreeNode;
    function MakeFilenames: boolean;
    procedure LoadReportDFM(src: string);
    procedure FixQR(aList: TStringList);
    procedure ClearTV(tv: TTreeview);
    procedure ReadIntoTree(aLines: TStrings);
    procedure FixNodes(tv: TTreeView);
    procedure GrabDataSources(tv: TTreeView);
    procedure ConvertReports(tv: TTreeView);
    procedure MoveBands(tv: TTreeView);
    procedure ConvertBands(tv: TTreeView);
    procedure FixBand(aList: TStringList);
    procedure ConvertQRPrintable(tv: TTreeView);
    procedure ConvertQRDBCalc(tv: TTreeView);
    procedure Tree2Memo(tv: TTreeView);
    procedure BuildDestDFM;
    procedure BuildDestPas;
    procedure ScrollTop(memo: TMemo);
    procedure ConvertDetailLink(tv: TTreeView);
    procedure ConvertImage(tv: TTreeView);
    procedure ConvertGroups(tv: TTreeView);
    procedure SetStatusText(val : string);
  public
    { Public declarations }
    property StatusText: string write SetStatusText;
  end;

  // Support functions
  function LineEndsWith(substr, str: string; MatchCase: boolean): boolean;
  function NameFromObjectText(str: string): string;
  procedure NameAndValue(var strName: string; var strValue: string; strSrc: string);
  procedure ListToMemo(aList: TStringList; aMemo: TMemo);

var
  frmConvert: TfrmConvert;

implementation

uses registry;

{$R *.DFM}
procedure ListToMemo(aList: TStringList; aMemo: TMemo);
begin
  if length(aList.Text) > 64000 then begin
    aMemo.Lines.Text := copy(aList.Text, 1, 64000);
    aMemo.Lines.Add('Display truncated to fit TMemo contrainsts');
  end
  else
    aMemo.Lines.Assign(aList);
end;

procedure NameAndValue(var strName: string; var strValue: string; strSrc: string);
var i : integer;
begin
  i := pos('=', strSrc);
  strName := trim(copy(strSrc, 1, i-1));
  strValue := trim(copy(strSrc, i+1, 99));
end;

function NameFromObjectText(str: string): string;
begin
  str := trim(copy(str, 8, 99));
  result := copy(str, 1, pos(':', str)-1);
end;

function LineEndsWith(substr, str: string; MatchCase: boolean) : boolean;
var
  sublen, len: integer;
begin
  result := false;
  sublen := length(substr);
  len := length(str);
  if sublen <= len then begin
    try
      str := copy(str, len - sublen + 1, sublen);
    except
      str := '';
    end;
    if MatchCase then
      result := substr = str
    else
      result := AnsiCompareStr(str, substr) = 0;
  end;
end;

constructor TQRConvert.Create;
begin
  srcDFM := tstringlist.create;
  srcPAS := tstringlist.create;
  DestDFM := tstringlist.create;
  DestPAS := tstringlist.create;
end;

destructor TQRConvert.Destroy;
begin
  srcDFM.Free;
  srcPAS.Free;
  DestDFM.Free;
  DestPAS.Free;
  inherited;
end;


procedure TfrmConvert.FormCreate(Sender: TObject);
var
  reg: TRegistry;
begin
  ProgressBar.Parent := StatusBar;

  QRConvert := TQRConvert.Create;

  QRDataSources := TStringList.Create;
  QRSubDetails := TStringList.Create;
  QRImages := TStringList.Create;
  QRHeaderBands := TStringList.Create;
  reg := TRegistry.Create;
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey(STR_APP_KEY, True) then begin
      SrcEdit.Text := reg.ReadString(STR_SRC_KEY);
      DestEdit.Text := reg.ReadString(STR_DEST_KEY);
      chkText.Checked := reg.ReadString(STR_DFM_KEY) = 'YES';
    end;
  finally
    Reg.CloseKey;
    Reg.Free;
  end;
end;

procedure TfrmConvert.FormClose(Sender: TObject; var Action: TCloseAction);
var
  reg: TRegistry;
begin
  QRConvert.Free;
  QRDataSources.Free;
  QRSubDetails.Free;
  QRImages.Free;
  QRHeaderBands.Free;
  reg := TRegistry.Create;
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey(STR_APP_KEY, True) then begin
      reg.WriteString(STR_SRC_KEY, SrcEdit.Text);
      reg.WriteString(STR_DEST_KEY, DestEdit.Text);
      if chkText.Checked then
        reg.WriteString(STR_DFM_KEY, 'YES')
      else
        reg.WriteString(STR_DFM_KEY, 'NO');
    end;
  finally
    Reg.CloseKey;
    Reg.Free;
  end;
end;

procedure TfrmConvert.SetStatusText(val : string);
begin
  StatusBar.Panels[0].Text := val;
  StatusBar.Repaint;
end;

procedure TfrmConvert.btnConvertClick(Sender: TObject);
begin
  btnConvert.Enabled := false;
  try
    if MakeFilenames then begin
      QRDataSources.Clear;
      QRSubDetails.Clear;
      QRImages.Clear;
      QRHeaderBands.Clear;

      ProgressBar.Position := 5;

      StatusText := 'Reading ' + SrcDFM;

      LoadReportDFM(SrcDFM);

      ProgressBar.Position := 10;

      if FileExists(SrcPAS) then begin
        StatusText := 'Reading ' + SrcPas;
        QRConvert.SrcPas.LoadFromFile(SrcPAS);

        ProgressBar.Position := 15;
        ListToMemo(QRConvert.SrcPas, memoSrcPas);

        ProgressBar.Position := 20;
        StatusText := 'Converting ' + SrcDFM;
        BuildDestDFM;

        ProgressBar.Position := 50;
        StatusText := 'Converting ' + SrcPas;
        BuildDestPas;

        ProgressBar.Position := 80;
        ScrollTop(memoDestPas);
        StatusText := 'Saving ' + SrcPas;
        QRConvert.DestPas.SaveToFile(DestPas);
        ProgressBar.Position := 100;
      end;
    end;
  finally
    ProgressBar.Position := 0;
    StatusText := 'Ready';
    btnConvert.Enabled := true;
  end;
end;

procedure TfrmConvert.BuildDestPas;
  function UnitName: string;
  begin
    result := ExtractFileName(DestPAS);
    result := copy(result, 1, pos('.', result)-1);
  end;

  // Fix the interface section of the code.  Replace the
  // unit name and replace the QR1 lib files with QR3 ones
  procedure FixInterface;
  var
    i : integer;
    str : string;
    InUses: boolean;
  begin
    InUses := false;
    for i := 0 to pred(QRConvert.DestPas.Count) do begin
      str := QRConvert.DestPas[i];
      if copy(str, 1, 4) = 'unit' then
        str := 'unit ' + UnitName + ';';
      if copy(str, 1, 4) = 'uses' then
        InUses := true;
      if str = 'type' then
        InUses := false;
      if str = 'implementation' then
        break;
      if InUses then
        str := StringReplace(str, 'Quickrep', 'QuickRpt, QRExpr, Qrctrls', [rfIgnoreCase]);
      QRConvert.DestPas[i] := str;
    end;
  end;

  // Change the component types using the lists built when we
  // converted the .dfm file
  procedure FixType;
  var
    i,j, TypeStart, TypeEnd : integer;
    str : string;
    InType: boolean;
  begin
    InType := false;
    TypeStart := 0;
    TypeEnd := 0;
    for i := 0 to pred(QRConvert.DestPas.Count) do begin
      str := QRConvert.DestPas[i];
      if copy(str, 1, 4) = 'type' then begin
        InType := true;
        TypeStart := i;
      end;
      if str = 'implementation' then begin
        TypeEnd := i;
        break;
      end;

      if InType then begin
        str := StringReplace(str, 'TQuickReport;', 'TQuickRep;', [rfIgnoreCase]);
        str := StringReplace(str, 'TQRDBCalc;', 'TQRExpr;', [rfIgnoreCase]);
        for j := 0 to pred(QRSubDetails.Count) do begin
          if pos(' ' + QRSubDetails[j] + ':', str) > 0 then begin
            str := StringReplace(str, 'TQRBand;', 'TQRSubDetail;', [rfIgnoreCase]);
            break;
          end;
        end;

        // Go the image stringlist and change all of the TImages used by
        // the report to TQRImages
        for j := 0 to pred(QRImages.Count) do begin
          if pos(' ' + QRImages[j] + ':', str) > 0 then begin
            str := StringReplace(str, 'TImage;', 'TQRImage;', [rfIgnoreCase]);
            break;
          end;
        end;
        // Flag the QR1 GroupHeader bands to be deleted.  The fields
        // on these bands were moved to the TQRGroup bands
        for j := 0 to pred(QRHeaderBands.Count) do begin
          if pos(' ' + QRHeaderBands[j] + ':', str) > 0 then begin
            str := STR_DELETE_MARKER + str;
            break;
          end;
        end;

        // Flag the TQRDetailLink components to be deleted
        if LineEndsWith('TQRDetailLink;', str, false) then
          str := STR_DELETE_MARKER + str;

        QRConvert.DestPas[i] := str;
      end;
    end;
    i := TypeEnd;

    // Remove the lines marked as deletable
    while i > TypeStart do begin
      if copy(QRConvert.DestPas[i], 1, length(STR_DELETE_MARKER)) = STR_DELETE_MARKER then
        QRConvert.DestPas.Delete(i);
      dec(i);
    end;
  end;

  // Change the event parameters from the QR1 values to the QR3 ones.
  procedure FixEvents;
  const
    str_bp = 'BeforePrint(';
  var
    i, j, k : integer;
    str, tmpstr, OldParam, NewParam : string;
  begin
    str := QRConvert.DestPas.Text;
    tmpstr := str;
    str := StringReplace(str, 'OpenString', 'String', [rfIgnoreCase, rfReplaceAll]);
    for i := low(QR_Events) to high(QR_Events) do begin
      if QR_Events[i].EventName <> '' then begin
        OldParam := QR_Events[i].EventName + QR_Events[i].Before;
        NewParam := QR_Events[i].EventName + QR_Events[i].After;

        str := StringReplace(str, OldParam, NewParam, [rfIgnoreCase, rfReplaceAll]);
      end;
    end;
    if str <> tmpstr then
      QRConvert.DestPas.Text := str;

    // Clean up
    k := length(str_bp);
    for i := 0 to pred(QRConvert.DestPas.Count) do begin
      str := TrimRight(QRConvert.DestPas[i]);
      j := length(str);
      tmpstr := copy(str, j - k + 1, k);
      if tmpstr = str_bp then
         QRConvert.DestPas[i] := str + 'Sender: TCustomQuickRep;';
    end;
  end;
begin
  QRConvert.DestPas.Assign(QRConvert.SrcPas);

  ListToMemo(QRConvert.DestPas, memoDestPas);

  FixInterface;
  FixType;
  FixEvents;
end;

function TfrmConvert.MakeFilenames: boolean;
begin
  result := true;
  SrcPAS := ChangeFileExt(SrcEdit.Text, '.pas');
  SrcDFM := ChangeFileExt(SrcEdit.Text, '.dfm');
  DestPAS := ChangeFileExt(DestEdit.Text, '.pas');
  DestDFM := ChangeFileExt(DestEdit.Text, '.dfm');

  // Simple check to prevent the user from overwriting the report
  if (SrcPas = DestPas) or (SrcDFM = DestDFM) then begin
    result := false;
    MessageDlg('Source and destination filenames can not be the same!', mtError, [mbOK], 0);
  end
  else begin
    if FileExists(DestPas) then begin
      result :=  MessageDlg(format(STR_FILE_EXISTS, [DestPas]),
        mtWarning, [mbOK, mbCancel], 0) = mrOk;
    end;
  end
end;

// Read the DFM file and convert it to text.
procedure TfrmConvert.LoadReportDFM(src: string);
var
  FileIn: TFileStream;
  TempStream : TMemoryStream;
  buffer: array[0..4] of char;
begin
  FileIn := TFileStream.Create(SrcDFM, fmOpenRead );

  // Read the first couple of bytes to see if we are a binary
  // dfm files or one converted to text
  FileIn.ReadBuffer(buffer, 2);
  FileIn.Seek(0, soFromBeginning);

  if buffer[0] <> 'o' then begin
    TempStream := TMemoryStream.Create;

    ObjectResourceToText(FileIn, TempStream);
    TempStream.Seek(0, soFromBeginning);
    QRConvert.SrcDFM.LoadFromStream(TempStream);
    TempStream.Free;
  end
  else begin
    QRConvert.SrcDFM.LoadFromStream(FileIn);
  end;

  ListToMemo(QRConvert.SrcDFM, memoSrcDFM);

  FileIn.Free;
end;

procedure TfrmConvert.ClearTV(tv: TTreeview);
var
  i: integer;
begin
  for i := 0 to tv.Items.Count-1 do
  begin
    if tv.Items[i].Data <> nil then
      TStringlist(tv.Items[i].Data).Free;
  end;
  tv.items.clear;
end;

// Clean up the stringlists for each object
procedure TfrmConvert.FixNodes(tv: TTreeView);
var
  i: integer;
  procedure FixNodeText(Node: TTreenode);
  var
    ii: integer;
    slist: TStringList;
  begin
    // first strip off all 'end' lines
    slist := TStringList(Node.Data);
    if slist <> nil then begin
      for ii := pred(slist.count) downto 0 do begin
        if trim(slist[ii]) = 'end' then
          slist.delete(ii)
        else
          break;
      end;
    end;
    // add the final end
    slist.Add('end');
    slist[0] := trim(slist[0]);
    for ii := 1 to slist.count-2 do
      slist[ii] := '  ' + trim(slist[ii]);
  end;
begin
  for i := 0 to pred(tv.Items.Count) do
    FixNodeText(tv.Items[i]);
end;

// We read each object from the dfm file and put it in a treeview.
// This serves two purposes one.  It provides a visible display of how
// the QR1 dfm file looks after it's converted to Qr3 and it provides
// a storage structure to allow us to manipulate the objects.
procedure TfrmConvert.ReadIntoTree(aLines: TStrings);
var
  len,
  i, j: integer;
  s1, s2, s3: string;
  oName, oType: string;
  ParentNode, NewNode: TTreeNode;
  CurrentIndent: integer;
  CurrentObjectText : TStringList;

  // A quick way to see where in the tree the node will go
  function GetIndent(str: string): integer;
  var ll: integer;
  begin
    result := 0;
    ll := pos('object', str);
    if ll > 0 then
      result := ll div 2;
  end;

  // We classify the object nodes by QR type.  This makes it
  // easy to go after each object by type
  function TypeToConst(str: string): integer;
  begin
    result := 0;
    if str = 'TQRBand' then result := QR_Band;
    if str = 'TQRShape' then result := QR_Printable;
    if str = 'TQRSysData' then result := QR_Printable;
    if str = 'TQRDBText' then result := QR_Printable;
    if str = 'TQRLabel' then result := QR_Printable;
    if str = 'TQRMemo' then result := QR_Printable;
    if str = 'TQRDBCalc' then result := QR_DBCALC;
    if str = 'TQuickReport' then result := QR_Report;
    if str = 'TDataSource' then result := QR_Datasource;
    if str = 'TTable' then result := QR_dataset;
    if str = 'TQuery' then result := QR_dataset;
    if str = 'TQRDetailLink' then result := QR_DetailLink;
    if str = 'TImage' then result := QR_Image;
    if str = 'TQRGroup' then result := qr_group;
  end;
begin
  len := length(STR_OBJECT);
  Memo2.Lines.Clear;
  ClearTV(tvQR);
  i := 0;
  CurrentObjectText := nil;
  NewNode := nil;
  ParentNode := nil;
  while i < alines.count do begin
    s1 := aLines[i];
    s2 := uppercase(trim(s1));

    // Take each object definition and make a new node in the
    // treeview.  The object properties get saved as a stringlist
    // attached to the node.
    if copy(s2, 1, len) = STR_OBJECT then begin
      j := pos(STR_OBJECT, uppercase(s1));
      s2 := copy(s1, j, len) + '-';
      s3 := copy(s1, j+len, 99);

      oName := copy(s1, j+len, 99);
      oName := copy(oName, 1, pos(':', oName)-1);
      oType := copy(s1, pos(':', s1)+2, 99);

      StatusText := 'Reading object ' + oName;

      CurrentIndent := GetIndent(s1);
      CurrentObjectText := TStringList.Create;

      if tvQR.Items.Count = 0 then begin
        ParentNode := tvQR.Items.AddObject(nil, oName, nil);
        ParentNode.Data := CurrentObjectText;
      end
      else begin
        if (CurrentIndent - 1) > ParentNode.Level then
          ParentNode := NewNode
        else if (CurrentIndent - 1) < ParentNode.Level then
          ParentNode := ParentNode.Parent;

        if ParentNode = nil then
          ParentNode := tvQR.Items[0];

        NewNode := tvQR.Items.AddChildObject(ParentNode, oName, nil);

        NewNode.Data := CurrentObjectText;

        NewNode.StateIndex := TypeToConst(oType);
      end;
    end;
    if CurrentObjectText <> nil then
      CurrentObjectText.Add(s1);
    inc(i);
  end;
  ProgressBar.Position := 25;
  // Clean up the object definition in the stringlist attached to each node
  FixNodes(tvQR);

  // Build a look up list of datasources and their datasets
  GrabDataSources(tvQR);

  StatusText := 'Converting QR1 objects to QR3 ';

  // Convert the TQuickReport class to a TQuickRep and make it the owner
  // of the bands
  ConvertReports(tvQR);

  ProgressBar.Position := 30;

  MoveBands(tvQR);

  ProgressBar.Position := 35;

  // remove or rename the QR1 properties from each band
  ConvertBands(tvQR);

  // remove or rename the QR1 properties from each QR component
  ConvertQRPrintable(tvQR);

  // Convert all TQRDBCalc components to TQRExpr
  ConvertQRDBCalc(tvQR);

  ProgressBar.Position := 40;

  // Convert all TImages on bands to TQRImages
  ConvertImage(tvQR);

  ProgressBar.Position := 45;
  
  // Convert TQRDetailLink to TQRSubDetail
  ConvertDetailLink(tvQR);

  // Convert QR1 Group bands to QR3 format
  ConvertGroups(tvQR);

  tvQR.FullExpand;
end;

procedure TfrmConvert.ConvertImage(tv: TTreeView);
var
  sList: TStringList;
  Node: TTreeNode;
begin
  Node := tv.Items[1];

  // Look for any TImage that is on a band and change it's type to TQRImage
  // and add it the QRImages list so we can fix the .pas file
  while Node <> nil do begin
    if (Node.StateIndex = QR_Image) and (Node.Parent.StateIndex = QR_Band) then begin
      sList := TStringList(Node.Data);
      if (sList <> nil) and (sList.Count > 1) then begin
        QRImages.Add(NameFromObjectText(sList[0]));
        slist[0] := StringReplace(slist[0], 'TImage', 'TQRImage', []);
      end;
    end;
    Node := Node.GetNext;
  end;
end;

// Each QRDetailLink component gets converted to a QRSubDetail band
procedure TfrmConvert.ConvertDetailLink(tv: TTreeView);
var
  sList: TStringList;
  Node, PrevNode: TTreeNode;
  i: integer;
  tmp, nam, val: string;
  DetailBand,
  DataSet,
  Master,
  PrintBefore: string;
  procedure UpdateSubDetailBand;
  var
    LinkNode: TTreeNode;
    ii: integer;
  begin
    LinkNode := tv.Items[1];
    while LinkNode <> nil do begin
      if LinkNode.StateIndex = QR_Band then begin
        sList := TStringList(LinkNode.Data);
        if (sList <> nil) and (sList.Count > 1) then begin
          if NameFromObjectText(sList[0]) = DetailBand then begin
            QRSubDetails.Add(DetailBand);
            slist[0] := 'object ' + DetailBand + ': TQRSubDetail';
            for ii := slist.count-2 downto 1 do begin
              tmp := slist[ii];
              NameAndValue(Nam, Val, tmp);
              if nam = 'BandType' then slist.Delete(ii);
            end;
            slist.Insert(slist.count-1, '  Master = ' + Master);
            if DataSet <> '' then
              slist.Insert(slist.count-1, '  DataSet = ' + Dataset);
            slist.Insert(slist.count-1, '  PrintBefore = ' + PrintBefore);
            break;
          end;
        end;
      end;
      LinkNode := LinkNode.GetNext;
    end;
  end;

begin
  Node := tv.Items[1];

  while Node <> nil do begin
    // get the detail band
    // Get the dataset
    // Get the master
    // Get PrintBefore
    if Node.StateIndex = QR_DetailLink then begin
      sList := TStringList(Node.Data);
      if (sList <> nil) and (sList.Count > 1) then begin
        for i := slist.count-2 downto 1 do begin
          tmp := slist[i];
          NameAndValue(Nam, Val, tmp);
          if nam = 'DataSource' then DataSet := QRDataSources.Values[val];
          if nam = 'DetailBand' then DetailBand := val;
          if nam = 'Master' then Master := val;
          if nam = 'PrintBefore' then PrintBefore := val;
        end;
        UpdateSubDetailBand;
      end;
    end;
    PrevNode := Node;
    Node := Node.GetNext;
    if PrevNode.StateIndex = QR_DetailLink then
      PrevNode.Delete;
  end;
end;

procedure TfrmConvert.ConvertGroups(tv: TTreeView);
var
  sList, hList: TStringList;
  Node: TTreeNode;
  i, Level: integer;
  Expression, BeforePrint, AfterPrint, Height,
  tmp, nam, val: string;
  SortList: TStringList;

  procedure MoveToGroup;
  var
    PrevNode, HeaderNode: TTreeNode;
    hNam, hVal: string;
    ii: integer;
  begin
    HeaderNode := tv.Items[1];

    while HeaderNode <> nil do begin
      PrevNode := nil;
      if HeaderNode.Text = val then begin
        while HeaderNode.GetFirstChild <> nil do begin
          HeaderNode.GetFirstChild.MoveTo(Node, naAddChild);
        end;
        PrevNode := HeaderNode;
        hList := TStringList(HeaderNode.Data);
        if (hList <> nil) and (hList.Count > 1) then begin
          for ii := hlist.count-2 downto 1 do begin
            tmp := hlist[ii];
            NameAndValue(hNam, hVal, tmp);
            if hNam = 'BeforePrint' then BeforePrint := hVal;
            if hNam = 'AfterPrint' then AfterPrint := hVal;
            if hNam = 'Height' then Height := hVal;
          end;
        end;
        QRHeaderBands.Add(HeaderNode.Text);
      end;

      HeaderNode := HeaderNode.GetNext;
      if PrevNode <> nil then
        PrevNode.Delete;
    end;
  end;
begin
  Node := tv.Items[1];
  SortList := TStringList.Create;
  Level := 0;

  while Node <> nil do begin
    if Node.StateIndex = QR_Group then begin
      sList := TStringList(Node.Data);
      if (sList <> nil) and (sList.Count > 1) then begin
        BeforePrint := '';
        AfterPrint := '';
        Height := '';
        for i := slist.count-2 downto 1 do begin
          tmp := slist[i];
          NameAndValue(Nam, Val, tmp);
          if nam = 'Level' then begin
            try
              Level := StrToInt(val) + 1;
            except
              Level := 1;
            end;
            slist[i] := format('  Master = %s', [ReportNode.Text]);
          end;
          if nam = 'HeaderBand' then begin
            MoveToGroup;
            slist.Delete(i);
          end;
          if nam = 'OnNeedData' then slist.Delete(i);
          if nam = 'DataField' then begin
            Expression := copy(val,2,length(val)-2);
            slist.Delete(i);
          end;
          if nam = 'DataSource' then begin
            nam := 'Dataset';
            val := QRDataSources.Values[val];
            if val <> '' then
              Expression := val + '.' + expression;
            slist[i] := format('  Expression = ''%s''', [expression]);
          end;
        end;
        SortList.AddObject(format('%3d', [Level]), Node);
        if BeforePrint <> '' then slist.insert(1, '  BeforePrint = ' + BeforePrint);
        if AfterPrint <> '' then slist.insert(1, '  AfterPrint = ' + AfterPrint);
        if Height <> '' then slist.insert(1, '  Height = ' + Height);
      end;
    end;
    Node := Node.GetNext;
  end;

  // Now sort the groups by level
  for i := 0 to pred(SortList.Count) do
    TTreeNode(SortList.Objects[i]).MoveTo(ReportNode, naAddChild);

  SortList.Free;
end;

procedure TfrmConvert.ConvertQRDBCalc(tv: TTreeView);
var
  sList: TStringList;
  Node: TTreeNode;
  i: integer;
  Expression, operation,
  tmp, nam, val: string;
begin
  Node := tv.Items[1];

  while Node <> nil do begin
    if Node.StateIndex = QR_DBCALC then begin
      sList := TStringList(Node.Data);
      if (sList <> nil) and (sList.Count > 1) then begin
        slist[0] := StringReplace(slist[0], 'TQRDBCalc', 'TQRExpr', []);
        for i := slist.count-2 downto 1 do begin
          tmp := slist[i];
          NameAndValue(Nam, Val, tmp);
          if nam = 'Operation' then begin
            operation := val;
            slist.Delete(i);
          end;
          if nam = 'DataField' then begin
            Expression := copy(val,2,length(val)-2);
            slist.Delete(i);
          end;
          if nam = 'DataSource' then begin
            nam := 'Dataset';
            val := QRDataSources.Values[val];
            if val <> '' then
              Expression := val + '.' + expression;
            if operation = 'qrcCOUNT' then Expression := 'COUNT';
            if operation = 'qrcSUM' then Expression := 'SUM(' + expression + ')';
            if operation = 'qrcMIN' then Expression := 'MIN(' + expression + ')';
            if operation = 'qrcMAX' then Expression := 'MAX(' + expression + ')';
            if operation = 'qrcAVERAGE' then Expression := 'AVE(' + expression + ')';
            slist[i] := format('  expression = ''%s''', [expression]);
          end;
        end;
      end;
    end;
    Node := Node.GetNext;
  end;
end;

procedure TfrmConvert.ConvertQRPrintable(tv: TTreeView);
var
  sList: TStringList;
  Node: TTreeNode;
  i: integer;
  tmp, nam, val: string;
  AutoSizeFixed,
  RightAligned : boolean;
begin
  Node := tv.Items[1];

  while Node <> nil do begin
    if Node.StateIndex = QR_Printable then begin
      sList := TStringList(Node.Data);
      if (sList <> nil) and (sList.Count > 1) then begin
        RightAligned := false;
        for i := slist.count-2 downto 1 do begin
          tmp := slist[i];
          NameAndValue(Nam, Val, tmp);
          // The QR3 components do not have a visual property.  You
          // can achieve the same affect by setting the enabled
          // property to false
          if nam = 'Visible' then begin
            slist[i] := format('  Enabled = %s', [val])
          end;
          if val = 'taRightJustify' then
            RightAligned := true;
          if nam = 'DataSource' then begin
            nam := 'Dataset';
            val := QRDataSources.Values[val];
            if val <> '' then
              slist[i] := format('  %s = %s', [nam, val])
            else
              slist.delete(i);
          end;
          if nam = 'TabStop' then
            slist.delete(i);
        end;
        // Right aligned text controls will have AutoSize forced to
        // false.  Otherwise the components will have odd Left property
        // values.
        if RightAligned then begin
          AutoSizeFixed := false;
          for i := slist.count-2 downto 1 do begin
            tmp := slist[i];
            NameAndValue(Nam, Val, tmp);
            if nam = 'AutoSize' then begin
              AutoSizeFixed := true;
              slist[i] := '  AutoSize = False';
              break;
            end;
          end;
          if not AutoSizeFixed then
            slist.Insert(1, '  AutoSize = False');
        end;
      end;
    end;
    Node := Node.GetNext;
  end;
end;

procedure TfrmConvert.FixBand(aList: TStringList);
var
  i: integer;
  tmp, nam, val: string;
  function GetQR3: boolean;
  var ii: integer;
  begin
    result := false;
    for ii := low(Band_Properties) to high(Band_Properties) do begin
      if Band_Properties[ii].Before = nam then begin
        nam := Band_Properties[ii].After;
        result := true;
        break;
      end;
    end;
  end;
begin
  for i := alist.count-2 downto 1 do begin
    tmp := alist[i];
    NameAndValue(Nam, Val, tmp);
    if GetQR3 then begin
      alist[i] := format('  %s = %s', [nam, val]);
    end
    else
      alist.Delete(i);
  end;
end;

procedure TfrmConvert.ConvertBands(tv: TTreeView);
var
  sList: TStringList;
  Node: TTreeNode;
begin
  Node := tv.Items[1];

  while Node <> nil do begin
    if Node.StateIndex = QR_BAND then begin
      sList := TStringList(Node.Data);
      if (sList <> nil) and (sList.Count > 1) then begin
        FixBand(sList)
      end;
    end;
    Node := Node.GetNext;
  end;
end;

procedure TfrmConvert.MoveBands(tv: TTreeView);
var
  BandNode,
  Node: TTreeNode;
  function MoveableNode(aNode: TTreeNode): boolean;
  begin
    result := (aNode.StateIndex = QR_Band) or (aNode.StateIndex = QR_Group);
  end;
begin
  if ReportNode <> nil then begin
    // First make the ReportNode the last component
    ReportNode.MoveTo(tv.Items[0], naAddChild);

    Node := tv.Items[1];
    while Node <> ReportNode do begin
      BandNode := Node;

      if MoveableNode(BandNode) then begin
        Node := Node.GetNextSibling;
        if Node = nil then begin
          Node := Node.GetLastChild;
          if Node = nil then
            Node := BandNode.GetNext
          else
            Node := Node.GetNext;
        end;
      end
      else
        Node := Node.GetNext;

      if MoveableNode(BandNode) then
        BandNode.MoveTo(ReportNode, naAddChild);
    end;
  end;
end;

procedure TfrmConvert.ConvertReports(tv: TTreeView);
var
  sList: TStringList;
  Node: TTreeNode;
begin
  Node := tv.Items[0];

  ReportNode := nil;

  while Node <> nil do begin
    if Node.StateIndex = QR_Report then begin
      if ReportNode = nil then
        ReportNode := Node;
      sList := TStringList(Node.Data);
      if (sList <> nil) and (sList.Count > 1) then begin
        FixQR(sList)
      end;
    end;
    Node := Node.GetNext;
  end;
end;

procedure TfrmConvert.GrabDataSources(tv: TTreeView);
var
  tmpStr: string;
  sList: TStringList;
  Node: TTreeNode;
begin
  Node := tv.Items[0];

  QRDataSources.Clear;

  while Node <> nil do begin
    if Node.StateIndex = QR_Datasource then begin
      sList := TStringList(Node.Data);
      if (sList <> nil) and (sList.Count > 1) then begin
        tmpStr := sList[1];
        QRDataSources.Add(format('%s=%s', [Node.Text, copy(tmpStr, pos ('=', tmpStr)+2,99)]));
      end;
    end;
    Node := Node.GetNext;
  end;
end;

procedure TfrmConvert.FixQR(aList: TStringList);
var
  i: integer;
  tmp, nam, val: string;
  function GetQR3: boolean;
  var ii: integer;
  begin
    result := false;
    for ii := low(qr_Properties) to high(qr_Properties) do begin
      if qr_Properties[ii].Before = nam then begin
        nam := qr_Properties[ii].After;
        result := true;
        break;
      end;
    end;
  end;
  procedure CheckEvents;
  var ii: integer;
  begin
    for ii := low(QR_Events) to high(QR_Events) do begin
      if nam = QR_Events[ii].Event then
        QR_Events[ii].EventName := val;
    end;
  end;
begin
  if alist.count > 0 then
    alist[0] := StringReplace(alist[0], 'TQuickReport', 'TQuickRep', []);

  // First read the events
  for i := low(QR_Events) to high(QR_Events) do begin
    QR_Events[i].EventName := '';
  end;

  for i := alist.count-2 downto 1 do begin
    tmp := alist[i];
    NameAndValue(Nam, Val, tmp);

    CheckEvents;

    if GetQR3 then begin
      if nam = 'Dataset' then begin
        val := QRDataSources.Values[val];
      end;
      if nam = 'Page.PaperSize' then begin
        val := copy(val, 4, 99);
      end;

      if val <> '' then
        alist[i] := format('  %s = %s', [nam, val])
      else
        alist.Delete(i);
    end
    else
      alist.Delete(i);
  end;
end;

procedure TfrmConvert.tvQRChange(Sender: TObject; Node: TTreeNode);
begin
  if Node.Data <> nil then begin
    Memo2.Lines.Assign(TStringList(Node.Data));
  end;
end;

procedure TfrmConvert.ScrollTop(memo: TMemo);
begin
  Memo.Perform(EM_LineScroll, 0, Memo.Lines.Count * -1);
end;

procedure TfrmConvert.BuildDestDFM;
var
  FileOut: TFileStream;
  TempStream : TMemoryStream;
begin
  ReadIntoTree(QRConvert.SrcDFM);
  QRConvert.DestDFM.Clear;
  memoDestDfm.Lines.Clear;
  Tree2Memo(tvQR);
  PageControl1.ActivePage := PageControl1.Pages[2];

  if chkText.Checked then begin
    QRConvert.DestDFM.SaveToFile(DestDFM);
  end
  else begin
    FileOut := TFileStream.Create(DestDFM, fmOpenWrite);
    FileOut.Size := 0;
    TempStream := TMemoryStream.Create;

    QRConvert.DestDFM.SaveToStream(TempStream);
    TempStream.Seek(0, soFromBeginning);

    ObjectTextToResource(TempStream, FileOut);

    FileOut.Free;
    TempStream.Free;
  end;

  ScrollTop(memoDestDfm);
end;

// Here is where we write out the components from TTreeView back
// to the dfm file.
procedure TfrmConvert.Tree2Memo(tv: TTreeView);
const spaces = '                            ';
var
  Node: TTreeNode;

  procedure PrintNode(aNode: TTreeNode);
  var
    sList: TStringList;
    i: integer;
    space: string;
  begin
    sList := TStringList(aNode.Data);
    space := copy(spaces, 1, aNode.Level * 2);
    
    for i := 0 to slist.count-2 do
      QRConvert.DestDFM.Add(space + sList[i]);

    for i := 0 to (aNode.Count - 1) do
      PrintNode(aNode.Item[I]);

    QRConvert.DestDFM.Add(space + sList[slist.count-1]);
  end;
begin
  Node := tv.Items[0];

  PrintNode(Node);

  ListToMemo(QRConvert.DestDFM, memoDestDfm);
end;

procedure TfrmConvert.btnSrcClick(Sender: TObject);
begin
  with odSrc do begin
    InitialDir := ExtractFilePath(SrcEdit.Text);
    if Execute then
      SrcEdit.Text := FileName;
  end;
end;

procedure TfrmConvert.btnDestClick(Sender: TObject);
begin
  with sdDest do begin
    InitialDir := ExtractFilePath(DestEdit.Text);
    if Execute then
      DestEdit.Text := FileName;
  end;
end;

procedure TfrmConvert.StatusBarDrawPanel(StatusBar: TStatusBar;
  Panel: TStatusPanel; const Rect: TRect);
var
  aRect: TRect;
begin
  if Panel = StatusBar.Panels[1] then begin
    aRect := Rect;
    InflateRect(aRect, 1, 1);
    ProgressBar.BoundsRect := aRect;
  end;
end;

end.
